import numpy as np

from typing import List, Tuple, FrozenSet, Dict, Callable, Any, Optional, Protocol
from sklearn.feature_extraction.text import CountVectorizer, ENGLISH_STOP_WORDS
from joblib import Parallel, delayed, effective_n_jobs
from functools import reduce
from collections import Counter
from toponymy.utility_functions import diversify_max_alpha as diversify
from vectorizers.transformers import InformationWeightTransformer
from sklearn.metrics import pairwise_distances

from sentence_transformers import SentenceTransformer

import scipy.sparse
import numba

from tqdm.auto import tqdm

Ngrammer = Callable[[str], List[str]]

from typing import Union, overload, TypeVar, cast


# Define a protocol for objects that behave like Tokenizers
class TokenizerLike(Protocol):
    def encode(self, text: str, *args: Any, **kwargs: Any) -> Any: ...

    def decode(self, ids: Any, *args: Any, **kwargs: Any) -> str: ...


def create_tokenizers_ngrammer(
    tokenizer: TokenizerLike, ngram_range: Tuple[int, int] = (1, 4)
) -> Ngrammer:
    """
    Creates an ngrammer function that uses a tokenizer to tokenize the text and then generates n-grams.

    Parameters
    ----------
    tokenizer : TokenizerLike
        A tokenizer object that has encode and decode methods.
    ngram_range : Tuple[int, int], optional
        The range of n-grams to consider, by default (1, 4).

    Returns
    -------
    Ngrammer
        A function that takes a string and returns a list of n-grams.
    """

    def ngrammer(text: str) -> List[str]:
        encoded = tokenizer.encode(text)
        if isinstance(encoded, list):
            tokens = encoded
        else:
            tokens = encoded.ids
        return [
            tokenizer.decode(tokens[i : i + n])
            for n in range(ngram_range[0], ngram_range[1] + 1)
            for i in range(len(tokens) - n)
        ]

    return ngrammer


def count_docs_ngrams(
    docs: List[str],
    ngrammer: Ngrammer,
    stop_words: FrozenSet[str],
    max_ngrams: int = 250_000,
) -> Dict[str, int]:
    result = {}
    for doc in docs:
        for gram in ngrammer(doc):
            split_gram = gram.split()
            if len(split_gram) > 0 and (
                split_gram[0] in stop_words or split_gram[-1] in stop_words
            ):
                continue
            if gram in result:
                result[gram] += 1
            else:
                result[gram] = 1

    if len(result) > max_ngrams:
        trim_value = np.sort(list(result.values()))[-max_ngrams]
        result = {key: value for key, value in result.items() if value >= trim_value}

    return result


def combine_dicts(
    dict1: Dict[str, int], dict2: Dict[str, int], max_ngrams: int = 250_000
) -> Dict[str, int]:
    result = dict1
    for key in dict2:
        if key in result:
            result[key] += dict2[key]
        else:
            result[key] = dict2[key]

    if len(result) > max_ngrams:
        trim_value = np.sort(list(result.values()))[-max_ngrams]
        result = {key: value for key, value in result.items() if value >= trim_value}
    return result


def _combine_tree_layer(
    dict_list: List[Dict[str, int]], max_ngrams: int = 250_000
) -> List[Dict[str, int]]:
    result = []
    for i in range(0, len(dict_list) - 1, 2):
        result.append(
            combine_dicts(dict_list[i], dict_list[i + 1], max_ngrams=max_ngrams)
        )
    if len(dict_list) % 2 == 1:
        result.append(dict_list[-1])
    return result


def tree_combine_dicts(
    dict_list: List[Dict[str, int]], max_ngrams: int = 250_000
) -> Dict[str, int]:
    while len(dict_list) > 1:
        dict_list = _combine_tree_layer(dict_list, max_ngrams=max_ngrams)
    return dict_list[0]


def build_count_matrix(
    docs: List[str], vocab: Dict[str, int], ngrammer: Ngrammer
) -> scipy.sparse.csr_matrix:
    col_indices = []
    indptr = [0]
    data = []
    for doc in docs:
        ngram_counter = {}
        for gram in ngrammer(doc):
            try:
                ngram_idx = vocab[gram]
                if ngram_idx in ngram_counter:
                    ngram_counter[ngram_idx] += 1
                else:
                    ngram_counter[ngram_idx] = 1
            except KeyError:
                continue

        col_indices.extend(ngram_counter.keys())
        data.extend(ngram_counter.values())
        indptr.append(len(col_indices))

    col_indices = np.asarray(col_indices, dtype=np.int32)
    data = np.asarray(data, dtype=np.int32)
    indptr = np.asarray(indptr, dtype=np.int32)

    X = scipy.sparse.csr_matrix(
        (data, col_indices, indptr), shape=(len(docs), len(vocab))
    )
    X.sort_indices()
    return X


def build_keyphrase_vocabulary(
    objects: List[str],
    ngrammer: Ngrammer,
    max_features: int = 50_000,
    min_occurrences: int = 1,
    stop_words: FrozenSet[str] = ENGLISH_STOP_WORDS,
    n_jobs: int = -1,
    min_chunk_size: int = 20_000,
    verbose: bool = False,
) -> List[str]:
    """
    Builds a keyphrase vocabulary from a list of objects.

    Parameters
    ----------
    objects : List[str]
        A list of objects; for use in building avocabulary this should be string representations of the objects.
    ngrammer : Ngrammer
        A function that takes a string and returns a list of n-grams.
    max_features : int, optional
        The maximum number of features to consider, by default 50_000.
    min_occurrences : int, optional
        The minimum number of occurrences for a keyphrase to be included, by default 1.
    stop_words : FrozenSet[str], optional
        The set of stop words to use, by default sklearn.feature_extraction.text.ENGLISH_STOP_WORDS.
    n_jobs : int, optional
        The number of jobs to use in parallel processing, by default -1. If -1, all available cores are used.
    min_chunk_size : int, optional
        The minimum chunk size for parallel processing, by default 20_000.
    verbose : bool, optional
        Whether to print out progress information, by default False.

    Returns
    -------
    List[str]
        A keyphrase list of the most commonly occurring keyphrases.
    """
    # count ngrams in parallel with joblib
    n_chunks = effective_n_jobs(n_jobs)
    chunk_size = max((len(objects) // n_chunks) + 1, min_chunk_size)
    n_chunks = len(objects) // chunk_size + 1
    if verbose:
        print(
            f"Chunking into {n_chunks} chunks of size {chunk_size} for keyphrase identification."
        )
    chunked_count_dicts = Parallel(n_jobs=n_chunks)(
        delayed(count_docs_ngrams)(
            objects[i : i + chunk_size],
            ngrammer,
            stop_words,
            max_ngrams=max_features * 10,
        )
        for i in range(0, len(objects), chunk_size)
    )

    if verbose:
        print("Combining count dictionaries ...")
    # Combine dictionaries and count the most common ngrams
    # all_vocab_counts = reduce(combine_dicts, chunked_count_dicts, {})
    all_vocab_counts = tree_combine_dicts(
        chunked_count_dicts, max_ngrams=max_features * 10
    )
    vocab_counter = Counter(all_vocab_counts)
    result = [
        ngram
        for ngram, occurrences in vocab_counter.most_common(max_features)
        if occurrences >= min_occurrences
    ]
    if len(result) == 0:
        raise ValueError(
            "No keyphrases found. Try increasing the max_features parameter or check that there are any re-occuring sections of text."
        )

    return result


def build_keyphrase_count_matrix(
    objects: List[str],
    keyphrases: Dict[str, int],
    ngrammer: Ngrammer,
    n_jobs: int = -1,
    min_chunk_size: int = 20_000,
    verbose: bool = False,
) -> scipy.sparse.spmatrix:
    """
    Builds a count matrix of keyphrases in a list of objects.

    Parameters
    ----------
    objects : List[str]
        A list of objects; for use in building a count matrix this should be string representations of the objects.
    keyphrases : Dict[str, int]
        A dictionary where keys are keyphrases to count in the objects and values are their respective indices.
    ngrammer : Ngrammer
        A function that takes a string and returns a list of n-grams.
    n_jobs : int, optional
        The number of jobs to use in parallel processing, by default -1. If -1, all available cores are used.
    min_chunk_size : int, optional
        The minimum chunk size for parallel processing, by default 20_000.
   verbose : bool, optional
        Whether to print out progress information, by default False.


    Returns
    -------
    scipy.sparse.spmatrix
        A sparse count matrix of keyphrases in the objects.
    """
    # count ngrams in parallel with joblib
    n_chunks = effective_n_jobs(n_jobs)
    chunk_size = max((len(objects) // n_chunks) + 1, min_chunk_size)
    n_chunks = (len(objects) // chunk_size) + 1
    if verbose:
        print(
            f"Chunking into {n_chunks} chunks of size {chunk_size} for keyphrase count construction."
        )
    chunked_count_matrices = Parallel(n_jobs=n_chunks)(
        delayed(build_count_matrix)(objects[i : i + chunk_size], keyphrases, ngrammer)
        for i in range(0, len(objects), chunk_size)
    )
    if verbose:
        print("Combining count matrix chunks ...")

    # stack the count matrices
    result = scipy.sparse.vstack(chunked_count_matrices)

    return result


def build_object_x_keyphrase_matrix(
    objects: List[str],
    ngram_range: Tuple[int, int] = (1, 4),
    tokenizer: Optional[TokenizerLike] = None,
    token_pattern: str = "(?u)\\b\\w[-'\\w]+\\b",
    max_features: int = 50_000,
    min_occurrences: int = 1,
    stop_words: FrozenSet[str] = ENGLISH_STOP_WORDS,
    n_jobs: int = -1,
    min_chunk_size: int = 20_000,
    verbose: bool = False,
) -> scipy.sparse.spmatrix:
    """
    Builds a count matrix of keyphrases in a list of objects.

    Parameters
    ----------
    objects : List[str]
        A list of objects; for use in building a count matrix this should be string representations of the objects.
    ngram_range : Tuple[int, int], optional
        The range of n-grams to consider, by default (1, 4).
    tokenizer : Optional[TokenizerLike], optional
        A tokenizer object that has encode and decode methods, by default None. If None, a CountVectorizer is used.
    token_pattern : str, optional
        The regular expression pattern to use for tokenization, by default "(?u)\\b\\w[-'\\w]+\\b".
    max_features : int, optional
        The maximum number of features to consider, by default 50_000.
    min_occurrences : int, optional
        The minimum number of occurrences for a keyphrase to be included, by default 1.
    stop_words : FrozenSet[str], optional
        The set of stop words to use, by default sklearn.feature_extraction.text.ENGLISH_STOP_WORDS.
    n_jobs : int, optional
        The number of jobs to use in parallel processing, by default -1. If -1, all available cores are used.
    min_chunk_size : int, optional
        The minimum chunk size for parallel processing, by default 20_000.
    verbose : bool, optional
        Whether to print out progress information, by default False.

    Returns
    -------
    scipy.sparse.spmatrix
        A sparse count matrix of keyphrases in the objects.
    """
    if tokenizer is None:
        # use Countvectorizer to build an ngram analyzer to ensure compatiability
        cv = CountVectorizer(
            lowercase=True,
            token_pattern=token_pattern,
            ngram_range=ngram_range,
        )
        ngrammer = cv.build_analyzer()
    else:
        ngrammer = create_tokenizers_ngrammer(tokenizer, ngram_range=ngram_range)

    keyphrases = build_keyphrase_vocabulary(
        objects,
        ngrammer=ngrammer,
        max_features=max_features,
        min_occurrences=min_occurrences,
        stop_words=stop_words,
        n_jobs=n_jobs,
        verbose=verbose,
    )
    if verbose:
        print(f"Found {len(keyphrases)} keyphrases.")

    keyphrase_dict = {keyphrase: i for i, keyphrase in enumerate(keyphrases)}
    result = build_keyphrase_count_matrix(
        objects,
        keyphrase_dict,
        ngrammer=ngrammer,
        n_jobs=n_jobs,
        verbose=verbose,
    )

    return result, keyphrases


class KeyphraseBuilder:
    """
    A class for building keyphrase count matrices from a list of objects. This can be useful
    as keyphrases can be a more specific way of helping prompt an LLM for a topic name. To
    make use of keyphrases you need to be able to convert objects to text. For basic
    short-text topic modeling, you can use the default settings, which simply assumes
    objects are already short texts. For other kinds of topic modeling you may
    need to provide a function that converts objects to text.

    Parameters
    ----------
    object_to_text : Optional[Callable[[Any], str]], optional
        A function that converts objects to text, by default None. If None, it is assumed that the objects are strings.
        An example of another case would be if objects were images and this function was a
        zero-short image captioning model.

    ngram_range : Tuple[int, int], optional
        The range of n-grams to consider, by default (1, 4).

    tokenizer : Optional[TokenizerLike], optional
        A tokenizer object that has encode and decode methods, by default None. If None, a CountVectorizer is used.

    token_pattern : str, optional
        The regular expression pattern to use for tokenization, by default "(?u)\\b\\w[-'\\w]+\\b".

    max_features : int, optional
        The maximum number of features to consider, by default 50_000.

    min_occurrences : int, optional
        The minimum number of occurrences for a keyphrase to be included, by default 2, so keyphrases have to re-occur.

    stop_words : FrozenSet[str], optional
        The set of stop words to use, by default sklearn.feature_extraction.text.ENGLISH_STOP_WORDS.

    n_jobs : int, optional
        The number of jobs to use in parallel processing, by default -1. If -1, all available cores are used.

    Attributes
    ----------

    object_x_keyphrase_matrix_ : scipy.sparse.spmatrix
        A sparse count matrix of keyphrases in the objects.

    keyphrase_list_ : List[str]
        A list of keyphrases in the same order as columns in object_x_keyphrase_matrix.

    """

    def __init__(
        self,
        object_to_text: Optional[Callable[[Any], str]] = None,
        ngram_range: Tuple[int, int] = (1, 4),
        tokenizer: Optional[TokenizerLike] = None,
        token_pattern: str = "(?u)\\b\\w[-'\\w]+\\b",
        max_features: int = 50_000,
        min_occurrences: int = 2,
        stop_words: FrozenSet[str] = ENGLISH_STOP_WORDS,
        n_jobs: int = -1,
        embedder: Optional[SentenceTransformer] = None,
        verbose: bool = False,
    ):
        self.object_to_text = object_to_text
        self.ngram_range = ngram_range
        self.tokenizer = tokenizer
        self.token_pattern = token_pattern
        self.max_features = max_features
        self.min_occurrences = min_occurrences
        self.stop_words = stop_words
        self.n_jobs = n_jobs
        self.embedder = embedder
        self.verbose = verbose

    def fit(self, objects: List[Any]):
        if self.object_to_text is None:
            object_texts = objects
        else:
            object_texts = [self.object_to_text(obj) for obj in objects]

        if self.verbose:
            print("Building keyphrase matrix ... ")

        self.object_x_keyphrase_matrix_, self.keyphrase_list_ = (
            build_object_x_keyphrase_matrix(
                object_texts,
                ngram_range=self.ngram_range,
                tokenizer=self.tokenizer,
                token_pattern=self.token_pattern,
                max_features=self.max_features,
                min_occurrences=self.min_occurrences,
                stop_words=self.stop_words,
                n_jobs=self.n_jobs,
                verbose=self.verbose,
            )
        )
        
        if self.embedder is not None:
            if self.verbose:
                print("Building keyphrase vectors ... ")

            self.keyphrase_vectors_ = self.embedder.encode(
                self.keyphrase_list_, show_progress_bar=self.verbose,
            )
        else:
            self.keyphrase_vectors_ = None

        return self

    def fit_transform(self, objects: List[Any]) -> Tuple[scipy.sparse.spmatrix, List[str], Optional[np.ndarray]]:
        """
        Fits the KeyphraseBuilder to the objects and returns the object x keyphrase matrix, keyphrase list, and keyphrase vectors.
        """
        self.fit(objects)
        return self.object_x_keyphrase_matrix_, self.keyphrase_list_, self.keyphrase_vectors_


@numba.njit()
def longest_keyphrases(candidate_keyphrases):  # pragma: no cover
    """
    Builds a list of keyphrases that are not substrings of other keyphrases.
    """
    result = []
    for i, phrase in enumerate(candidate_keyphrases):
        for other in candidate_keyphrases:
            if f" {phrase}" in other or f"{phrase} " in other:
                phrase = other

        if phrase not in result:
            candidate_keyphrases[i] = phrase
            result.append(phrase)

    return result


def subset_matrix_and_class_labels(
    cluster_label_vector: np.ndarray,
    object_x_keyphrase_matrix: scipy.sparse.spmatrix,
) -> Tuple[scipy.sparse.spmatrix, np.ndarray, np.ndarray]:
    # Mask out noise points, and then columns and rows that then have no entries
    count_matrix = object_x_keyphrase_matrix[cluster_label_vector >= 0, :]
    column_mask = np.squeeze(np.asarray(count_matrix.sum(axis=0))) > 0.0
    count_matrix = count_matrix[:, column_mask]
    column_map = np.arange(object_x_keyphrase_matrix.shape[1])[column_mask]
    row_mask = np.squeeze(np.asarray(count_matrix.sum(axis=1))) > 0.0
    count_matrix = count_matrix[row_mask, :]

    # Make a label vector contracted to the appropriate space
    class_labels = cluster_label_vector[cluster_label_vector >= 0][row_mask]

    return count_matrix, class_labels, column_map


def information_weighted_keyphrases(
    cluster_label_vector: np.ndarray,
    object_x_keyphrase_matrix: scipy.sparse.spmatrix,
    keyphrase_list: List[str],
    keyphrase_vectors: np.ndarray,
    embedding_model: SentenceTransformer,
    n_keyphrases: int = 16,
    prior_strength: float = 0.1,
    weight_power: float = 2.0,
    diversify_alpha: float = 1.0,
    show_progress_bar: bool = False,
) -> List[List[str]]:
    """Generates a list of keyphrases for each cluster in a cluster layer.

    Parameters
    ----------
    cluster_label_vector : np.ndarray
        A vector of cluster labels for each object.
    object_x_keyphrase_matrix : scipy.sparse.spmatrix
        A sparse matrix of keyphrase counts for each object.
    keyphrase_list : List[str]
        A list of keyphrases in the same order as columns in object_x_keyphrase_matrix.
    keyphrase_vectors : np.ndarray
        An ndarray of keyphrase vectors in the same order as columns in object_x_keyphrase_matrix.
    embedding_model : SentenceTransformer
        A SentenceTransformer model for embedding keyphrases.
    n_keyphrases : int, optional
        The number of keyphrases to generate for each cluster, by default 16.
    prior_strength : float, optional
        The strength of the prior for the information weighting, by default 0.1.
    weight_power : float, optional
        The power to raise the information weights to, by default 2.0.
    diversify_alpha : float, optional
        The alpha parameter for diversifying the keyphrase selection, by default 1.0.
    show_progress_bar : bool, optional
        Whether to show a progress bar for the computation, by default False.

    Returns
    -------
    keyphrases List[List[str]]
        A list of lists of keyphrases for each cluster.
    """
    keyphrase_vector_mapping = {
        keyphrase: vector
        for keyphrase, vector in zip(keyphrase_list, keyphrase_vectors)
        if not np.all(vector == 0.0)
    }
    count_matrix, class_labels, column_map = subset_matrix_and_class_labels(
        cluster_label_vector, object_x_keyphrase_matrix
    )

    iwt = InformationWeightTransformer(
        prior_strength=prior_strength, weight_power=weight_power
    ).fit(count_matrix, class_labels)
    count_matrix.data = np.log(count_matrix.data + 1)
    count_matrix.eliminate_zeros()
    weighted_matrix = iwt.transform(count_matrix)

    result = []
    for cluster_num in tqdm(
        range(cluster_label_vector.max() + 1),
        desc="Generating informative keyphrases",
        disable=not show_progress_bar,
        leave=False,
        unit="cluster",
        position=1,
    ):
        # Sum over the cluster; get the top scoring indices
        contrastive_scores = np.squeeze(
            np.asarray(weighted_matrix[class_labels == cluster_num].sum(axis=0))
        )
        if sum(contrastive_scores) == 0:
            result.append(["No notable keyphrases"])
            continue

        keyphrases_present_indices = np.where(contrastive_scores > 0)[0]
        keyphrase_weights = contrastive_scores[keyphrases_present_indices]
        keyphrases_present = [
            keyphrase_list[column_map[j]] for j in keyphrases_present_indices
        ]
        # Update keyphrase mapping with present keyphrases it is missing
        missing_keyphrases = [
            keyphrase
            for keyphrase in keyphrases_present
            if keyphrase not in keyphrase_vector_mapping
        ]
        if len(missing_keyphrases) > 0:
            if embedding_model is None:
                raise ValueError(
                    "On demand keyphrase vectorization was requested but no embedding model provided. Please provide an embedding model."
                )
            missing_keyphrase_vectors = embedding_model.encode(
                missing_keyphrases, show_progress_bar=False
            )
            for keyphrase, vector in zip(missing_keyphrases, missing_keyphrase_vectors):
                keyphrase_vector_mapping[keyphrase] = vector

        # Compute the centroid of the keyphrases present in the cluster
        centroid_vector = np.average(
            [keyphrase_vector_mapping[keyphrase] for keyphrase in keyphrases_present],
            weights=keyphrase_weights,
            axis=0,
        )

        chosen_indices = np.argsort(contrastive_scores)[-max((n_keyphrases * 4), 16) :]

        # Map the indices back to the original vocabulary
        chosen_keyphrases = [
            keyphrase_list[column_map[j]]
            for j in reversed(chosen_indices)
            if j in keyphrases_present_indices
        ]

        # Extract the longest keyphrases, then diversify the selection
        chosen_keyphrases = longest_keyphrases(chosen_keyphrases)
        chosen_vectors = np.asarray(
            [keyphrase_vector_mapping[phrase] for phrase in chosen_keyphrases]
        )
        chosen_indices = diversify(
            centroid_vector,
            chosen_vectors,
            n_keyphrases,
            max_alpha=diversify_alpha,
        )[:n_keyphrases]
        chosen_keyphrases = [chosen_keyphrases[j] for j in chosen_indices]

        result.append(chosen_keyphrases)

    # Update keyphrase vectors with vectors from the mapping
    for i, keyphrase in enumerate(keyphrase_list):
        if keyphrase in keyphrase_vector_mapping:
            keyphrase_vectors[i] = keyphrase_vector_mapping[keyphrase]

    return result


def central_keyphrases(
    cluster_label_vector: np.ndarray,
    object_x_keyphrase_matrix: scipy.sparse.spmatrix,
    keyphrase_list: List[str],
    keyphrase_vectors: np.ndarray,
    embedding_model: SentenceTransformer,
    n_keyphrases: int = 16,
    diversify_alpha: float = 1.0,
    show_progress_bar: bool = False,
):
    """
    Generates a list of keyphrases for each cluster in a cluster layer using the central keyphrase method.

    Parameters
    ----------
    cluster_label_vector : np.ndarray
        A vector of cluster labels for each object.
    object_x_keyphrase_matrix : scipy.sparse.spmatrix
        A sparse matrix of keyphrase counts for each object.
    keyphrase_list : List[str]
        A list of keyphrases in the same order as columns in object_x_keyphrase_matrix.
    keyphrase_vectors : np.ndarray
        An ndarray of keyphrase vectors in the same order as columns in object_x_keyphrase_matrix.
    embedding_model : SentenceTransformer
        A SentenceTransformer model for embedding keyphrases.
    n_keyphrases : int, optional
        The number of keyphrases to generate for each cluster, by default 16.
    diversify_alpha : float, optional
        The alpha parameter for diversifying the keyphrase selection, by default 1.0.
    show_progress_bar : bool, optional
        Whether to show a progress bar for the computation, by default False.

    Returns
    -------
    keyphrases : List[List[str]]
        A list of lists of keyphrases for each cluster.
    """
    keyphrase_vector_mapping = {
        keyphrase: vector
        for keyphrase, vector in zip(keyphrase_list, keyphrase_vectors)
        if not np.all(vector == 0.0)
    }

    count_matrix, class_labels, column_map = subset_matrix_and_class_labels(
        cluster_label_vector, object_x_keyphrase_matrix
    )

    result = []
    for cluster_num in tqdm(
        range(cluster_label_vector.max() + 1),
        desc="Generating central keyphrases",
        disable=not show_progress_bar,
        leave=False,
        unit="cluster",
        position=1,
    ):
        # Sum over the cluster; get the non-zero indices
        base_candidate_indices = np.where(
            np.squeeze(
                np.asarray(count_matrix[class_labels == cluster_num].sum(axis=0))
            )
            > 0
        )[0]

        null_topic = np.mean(keyphrase_vectors, axis=0)

        # Map the indices back to the original vocabulary
        base_candidates = [
            keyphrase_list[column_map[j]] for j in base_candidate_indices
        ]
        # Update keyphrase mapping with present keyphrases it is missing
        missing_keyphrases = [
            keyphrase
            for keyphrase in base_candidates
            if keyphrase not in keyphrase_vector_mapping
        ]
        if len(missing_keyphrases) > 0:
            if embedding_model is None:
                raise ValueError(
                    "On demand keyphrase vectorization was requested but no embedding model provided. Please provide an embedding model."
                )
            missing_keyphrase_vectors = embedding_model.encode(
                missing_keyphrases, show_progress_bar=False
            )
            for keyphrase, vector in zip(missing_keyphrases, missing_keyphrase_vectors):
                keyphrase_vector_mapping[keyphrase] = vector

        base_vectors = np.asarray(
            [keyphrase_vector_mapping[phrase] for phrase in base_candidates]
        ) - null_topic
        base_weights = np.squeeze(np.asarray(count_matrix[class_labels == cluster_num].sum(axis=0)))[base_candidate_indices]
        centroid = np.average(base_vectors, axis=0, weights=base_weights)

        # Select the central keyphrases as the closest samples to the centroid
        base_distances = pairwise_distances(
            centroid.reshape(1, -1), base_vectors, metric="cosine"
        )
        base_order = np.argsort(base_distances.flatten())

        chosen_keyphrases = [base_candidates[i] for i in base_order[: n_keyphrases**2]]

        # Extract the longest keyphrases, then diversify the selection
        chosen_keyphrases = longest_keyphrases(chosen_keyphrases)
        chosen_vectors = np.asarray(
            [keyphrase_vector_mapping[phrase] for phrase in chosen_keyphrases]
        )
        chosen_indices = diversify(
            centroid,
            chosen_vectors,
            n_keyphrases,
            max_alpha=diversify_alpha,
        )[:n_keyphrases]
        chosen_keyphrases = [chosen_keyphrases[j] for j in chosen_indices]

        result.append(chosen_keyphrases)

    # Update keyphrase vectors with vectors from the mapping
    for i, keyphrase in enumerate(keyphrase_list):
        if keyphrase in keyphrase_vector_mapping:
            keyphrase_vectors[i] = keyphrase_vector_mapping[keyphrase]

    return result


def bm25_keyphrases(
    cluster_label_vector: np.ndarray,
    object_x_keyphrase_matrix: scipy.sparse.spmatrix,
    keyphrase_list: List[str],
    keyphrase_vectors: np.ndarray,
    embedding_model: SentenceTransformer,
    n_keyphrases: int = 16,
    k1: float = 1.5,
    b: float = 0.75,
    diversify_alpha: float = 1.0,
    show_progress_bar: bool = False,
) -> List[List[str]]:
    """Generates a list of keyphrases for each cluster in a cluster layer using BM25 for scoring.

    Parameters
    ----------
    cluster_label_vector : np.ndarray
        A vector of cluster labels for each object.
    object_x_keyphrase_matrix : scipy.sparse.spmatrix
        A sparse matrix of keyphrase counts for each object.
    keyphrase_list : List[str]
        A list of keyphrases in the same order as columns in object_x_keyphrase_matrix.
    keyphrase_vectors : np.ndarray
        An ndarray of keyphrase vectors in the same order as columns in object_x_keyphrase_matrix.
    embedding_model : SentenceTransformer
        A SentenceTransformer model for embedding keyphrases.
    n_keyphrases : int, optional
        The number of keyphrases to generate for each cluster, by default 16.
    k1 : float, optional
        The k1 parameter for BM25, by default 1.5.
    b : float, optional
        The b parameter for BM25, by default 0.75.
    diversify_alpha : float, optional
        The alpha parameter for diversifying the keyphrase selection, by default 1.0.

    Returns
    -------
    keyphrases : List[List[str]]
        A list of lists of keyphrases for each cluster.
    """
    keyphrase_vector_mapping = {
        keyphrase: vector
        for keyphrase, vector in zip(keyphrase_list, keyphrase_vectors)
        if not np.all(vector == 0.0)
    }

    count_matrix, class_labels, column_map = subset_matrix_and_class_labels(
        cluster_label_vector, object_x_keyphrase_matrix
    )

    # Build a class based count matrix
    groupby_matrix = scipy.sparse.csr_matrix(
        (
            np.ones(class_labels.shape[0]),
            (class_labels, np.arange(class_labels.shape[0])),
        ),
        shape=(class_labels.max() + 1, class_labels.shape[0]),
    )
    class_count_matrix = groupby_matrix @ count_matrix

    # Compute BM25 scores for every entry in the matrix
    N = class_count_matrix.shape[0]
    df = (class_count_matrix > 0).sum(axis=0)
    idf = np.log(1 + (N - df + 0.5) / (df + 0.5))

    doc_lengths = count_matrix.sum(axis=1)
    avg_doc_length = doc_lengths.mean()

    for i in range(class_count_matrix.shape[0]):
        tf_array = class_count_matrix.data[
            class_count_matrix.indptr[i] : class_count_matrix.indptr[i + 1]
        ]
        tf_score = tf_array / (
            k1 * ((1 - b) + b * doc_lengths[i] / avg_doc_length) + tf_array
        )
        class_count_matrix.data[
            class_count_matrix.indptr[i] : class_count_matrix.indptr[i + 1]
        ] = tf_score

    bm25_matrix = class_count_matrix.multiply(idf).tocsr()

    # Select the top scoring keyphrases for each cluster based on BM25 scores for the cluster
    result = []
    for cluster_num in tqdm(
        range(cluster_label_vector.max() + 1),
        desc="Generating bm25 keyphrases",
        disable=not show_progress_bar,
        leave=False,
        unit="meta-cluster",
        position=1,
    ):
        # Sum over the cluster; get the top scoring indices
        contrastive_scores = bm25_matrix[cluster_num].toarray().flatten()
        if sum(contrastive_scores) == 0:
            result.append(["No notable keyphrases"])
            continue

        keyphrases_present_indices = np.where(contrastive_scores > 0)[0]
        keyphrase_weights = contrastive_scores[keyphrases_present_indices]
        keyphrases_present = [
            keyphrase_list[column_map[j]] for j in keyphrases_present_indices
        ]
        # Update keyphrase mapping with present keyphrases it is missing
        missing_keyphrases = [
            keyphrase
            for keyphrase in keyphrases_present
            if keyphrase not in keyphrase_vector_mapping
        ]
        if len(missing_keyphrases) > 0:
            if embedding_model is None:
                raise ValueError(
                    "On demand keyphrase vectorization was requested but no embedding model provided. Please provide an embedding model."
                )
            missing_keyphrase_vectors = embedding_model.encode(
                missing_keyphrases, show_progress_bar=False
            )
            for keyphrase, vector in zip(missing_keyphrases, missing_keyphrase_vectors):
                keyphrase_vector_mapping[keyphrase] = vector

        # Compute the centroid of the keyphrases present in the cluster
        centroid_vector = np.average(
            [keyphrase_vector_mapping[keyphrase] for keyphrase in keyphrases_present],
            weights=keyphrase_weights,
            axis=0,
        )

        chosen_indices = np.argsort(contrastive_scores)[-(n_keyphrases**2) :]

        # Map the indices back to the original vocabulary
        chosen_keyphrases = [
            keyphrase_list[column_map[j]]
            for j in reversed(chosen_indices)
            if j in keyphrases_present_indices
        ]

        # Extract the longest keyphrases, then diversify the selection
        chosen_keyphrases = longest_keyphrases(chosen_keyphrases)
        chosen_vectors = np.asarray(
            [keyphrase_vector_mapping[phrase] for phrase in chosen_keyphrases]
        )
        chosen_indices = diversify(
            centroid_vector,
            chosen_vectors,
            n_keyphrases,
            max_alpha=diversify_alpha,
        )[:n_keyphrases]
        chosen_keyphrases = [chosen_keyphrases[j] for j in chosen_indices]

        result.append(chosen_keyphrases)

    # Update keyphrase vectors with vectors from the mapping
    for i, keyphrase in enumerate(keyphrase_list):
        if keyphrase in keyphrase_vector_mapping:
            keyphrase_vectors[i] = keyphrase_vector_mapping[keyphrase]

    return result



from abc import ABC, abstractmethod
import numpy as np
import numba
from fast_hdbscan.cluster_trees import (
    mst_to_linkage_tree,
    condense_tree,
    extract_leaves,
    get_cluster_label_vector,
)
from fast_hdbscan.boruvka import parallel_boruvka
from fast_hdbscan.numba_kdtree import kdtree_to_numba
from sklearn.neighbors import KDTree
from typing import List, Tuple, Dict, Type, Any, Optional
from toponymy.cluster_layer import ClusterLayer, ClusterLayerText
from typing import List, Tuple, Dict, Type, Any, Optional, NewType

from sklearn.cluster import KMeans

ClusterTree = NewType("ClusterTree", Dict[Tuple[int, int], List[Tuple[int, int]]])


@numba.njit(cache=True)
def binary_search_for_n_clusters(
    uncondensed_tree, approx_n_clusters, n_samples
):  # pragma: no cover
    lower_bound_min_cluster_size = 2
    upper_bound_min_cluster_size = n_samples // 2
    mid_min_cluster_size = int(
        round((lower_bound_min_cluster_size + upper_bound_min_cluster_size) / 2.0)
    )
    min_n_clusters = 0

    upper_tree = condense_tree(uncondensed_tree, upper_bound_min_cluster_size)
    leaves = extract_leaves(upper_tree)
    upper_n_clusters = len(leaves)

    lower_tree = condense_tree(uncondensed_tree, lower_bound_min_cluster_size)
    leaves = extract_leaves(lower_tree)
    lower_n_clusters = len(leaves)

    while upper_bound_min_cluster_size - lower_bound_min_cluster_size > 1:
        mid_min_cluster_size = int(
            round((lower_bound_min_cluster_size + upper_bound_min_cluster_size) / 2.0)
        )
        if (
            mid_min_cluster_size == lower_bound_min_cluster_size
            or mid_min_cluster_size == upper_bound_min_cluster_size
        ):
            break
        mid_tree = condense_tree(uncondensed_tree, mid_min_cluster_size)
        leaves = extract_leaves(mid_tree)
        mid_n_clusters = len(leaves)

        if mid_n_clusters < approx_n_clusters:
            upper_bound_min_cluster_size = mid_min_cluster_size
            upper_n_clusters = mid_n_clusters
        elif mid_n_clusters >= approx_n_clusters:
            lower_bound_min_cluster_size = mid_min_cluster_size
            lower_n_clusters = mid_n_clusters

    if abs(lower_n_clusters - approx_n_clusters) < abs(
        upper_n_clusters - approx_n_clusters
    ):
        lower_tree = condense_tree(uncondensed_tree, lower_bound_min_cluster_size)
        leaves = extract_leaves(lower_tree)
        clusters = get_cluster_label_vector(lower_tree, leaves, 0.0, n_samples)
        return leaves, clusters
    elif abs(lower_n_clusters - approx_n_clusters) > abs(
        upper_n_clusters - approx_n_clusters
    ):
        upper_tree = condense_tree(uncondensed_tree, upper_bound_min_cluster_size)
        leaves = extract_leaves(upper_tree)
        clusters = get_cluster_label_vector(upper_tree, leaves, 0.0, n_samples)
        return leaves, clusters
    else:
        lower_tree = condense_tree(uncondensed_tree, lower_bound_min_cluster_size)
        lower_leaves = extract_leaves(lower_tree)
        lower_clusters = get_cluster_label_vector(
            lower_tree, lower_leaves, 0.0, n_samples
        )
        upper_tree = condense_tree(uncondensed_tree, upper_bound_min_cluster_size)
        upper_leaves = extract_leaves(upper_tree)
        upper_clusters = get_cluster_label_vector(
            upper_tree, upper_leaves, 0.0, n_samples
        )

        if np.sum(lower_clusters >= 0) > np.sum(upper_clusters >= 0):
            return lower_leaves, lower_clusters
        else:
            return upper_leaves, upper_clusters


def build_raw_cluster_layers(
    data: np.ndarray,
    *,
    min_clusters: int = 3,
    min_samples: int = 5,
    base_min_cluster_size: int = 10,
    base_n_clusters: Optional[int] = None,
    next_cluster_size_quantile: float = 0.8,
    verbose=False,
) -> List[np.ndarray]:
    """
    Build hierarchical cluster layers from raw data using a KDTree and Boruvka's algorithm.

    Parameters
    ----------
    data : np.ndarray
        The input data array of shape (n_samples, n_features).
    min_clusters : int, optional
        The minimum number of clusters to form in each layer, by default 3.
    min_samples : int, optional
        The minimum number of samples in a cluster, by default 5.
    base_min_cluster_size : int, optional
        The initial minimum cluster size, by default 10.
    base_n_clusters : Optional[int], optional
        The initial number of clusters, by default None. If None, base_min_cluster_size is used.
        If not None, this value will override base_min_cluster_size.
    next_cluster_size_quantile : float, optional
        The quantile to determine the next minimum cluster size, by default 0.8.

    Returns
    -------
    List[np.ndarray]
        A list of numpy arrays, each representing cluster labels for a layer.
    """
    n_samples = data.shape[0]
    cluster_layers = []
    min_cluster_size = base_min_cluster_size

    sklearn_tree = KDTree(data)
    numba_tree = kdtree_to_numba(sklearn_tree)
    edges, _, _ = parallel_boruvka(
        numba_tree, min_samples=min_cluster_size if min_samples is None else min_samples
    )
    sorted_mst = edges[np.argsort(edges.T[2])]
    uncondensed_tree = mst_to_linkage_tree(sorted_mst)
    if base_n_clusters is not None:
        leaves, clusters = binary_search_for_n_clusters(
            uncondensed_tree, base_n_clusters, n_samples=n_samples
        )
        cluster_sizes = np.bincount(clusters[clusters >= 0])
        min_cluster_size = np.min(cluster_sizes)
    else:
        new_tree = condense_tree(uncondensed_tree, base_min_cluster_size)
        leaves = extract_leaves(new_tree)
        clusters = get_cluster_label_vector(new_tree, leaves, 0.0, n_samples)

    n_clusters_in_layer = clusters.max() + 1

    while n_clusters_in_layer >= min_clusters:
        if verbose:
            print(f"Layer {len(cluster_layers)} found {n_clusters_in_layer} clusters")
        cluster_layers.append(clusters)
        cluster_sizes = np.bincount(clusters[clusters >= 0])
        next_min_cluster_size = int(
            np.quantile(cluster_sizes, next_cluster_size_quantile)
        )
        if next_min_cluster_size <= min_cluster_size + 1:
            break
        else:
            min_cluster_size = next_min_cluster_size
        new_tree = condense_tree(uncondensed_tree, min_cluster_size)
        leaves = extract_leaves(new_tree)
        clusters = get_cluster_label_vector(new_tree, leaves, 0.0, n_samples)
        n_clusters_in_layer = clusters.max() + 1

    return cluster_layers


@numba.njit(cache=True)
def _build_cluster_tree(labels: np.ndarray) -> List[Tuple[int, int, int, int]]:

    mapping = [(-1, -1, -1, -1) for _ in range(0)]
    found = [set([-1]) for _ in range(len(labels))]
    for upper_layer in range(1, labels.shape[0]):
        upper_layer_unique_labels = np.unique(labels[upper_layer])
        for lower_layer in range(upper_layer - 1, -1, -1):
            upper_cluster_order = np.argsort(labels[upper_layer])
            cluster_groups = np.split(
                labels[lower_layer][upper_cluster_order],
                np.cumsum(np.bincount(labels[upper_layer] + 1))[:-1],
            )
            # If there is no noise we are off by one, and need to drop the first cluster group
            if len(cluster_groups) > upper_layer_unique_labels.shape[0]:
                cluster_groups = cluster_groups[1:]
            for i, label in enumerate(upper_layer_unique_labels):
                if label >= 0:
                    for child in cluster_groups[i]:
                        if child >= 0 and child not in found[lower_layer]:
                            mapping.append((upper_layer, label, lower_layer, child))
                            found[lower_layer].add(child)

    for lower_layer in range(labels.shape[0] - 1, -1, -1):
        for child in range(labels[lower_layer].max() + 1):
            if child >= 0 and child not in found[lower_layer]:
                mapping.append((labels.shape[0], 0, lower_layer, child))

    return mapping


def build_cluster_tree(
    labels: List[np.ndarray],
) -> ClusterTree:
    """
    Builds a cluster tree from the given labels.

    Parameters
    ----------
    labels : List[np.ndarray]
        A list of numpy arrays where each array represents the labels of clusters at a specific layer.

    Returns
    -------
    ClusterTree
        A dictionary where the keys are tuples representing the parent cluster (layer, cluster index)
        and the values are lists of tuples representing the child clusters (layer, cluster index).
    """
    result = {}
    raw_mapping = _build_cluster_tree(np.vstack(labels))
    for parent_layer, parent_cluster, child_layer, child_cluster in raw_mapping:
        parent_name = (parent_layer, parent_cluster)
        if parent_name in result:
            result[parent_name].append((child_layer, child_cluster))
        else:
            result[parent_name] = [(child_layer, child_cluster)]
    return result


@numba.njit()
def centroids_from_labels(
    cluster_labels: np.ndarray, vector_data: np.ndarray
) -> np.ndarray: # pragma: no cover
    result = np.zeros((cluster_labels.max() + 1, vector_data.shape[1]))
    counts = np.zeros(cluster_labels.max() + 1)
    for i in range(cluster_labels.shape[0]):
        cluster_num = cluster_labels[i]
        if cluster_num >= 0:
            result[cluster_num] += vector_data[i]
            counts[cluster_num] += 1

    for i in range(result.shape[0]):
        if counts[i] > 0:
            result[i] /= counts[i]

    return result


def create_cluster_layers(
    layer_class: Type[Any],
    clusterable_vectors: np.ndarray,
    embedding_vectors: np.ndarray,
    min_clusters: int = 6,
    min_samples: int = 5,
    base_min_cluster_size: int = 10,
    base_n_clusters: Optional[int] = None,
    next_cluster_size_quantile: float = 0.8,
    show_progress_bar: bool = False,
    verbose: bool = False,
    **layer_kwargs,
) -> Tuple[List[ClusterLayer], ClusterTree]:
    """
    Create cluster layers from given vectors and parameters.

    Parameters
    ----------
    layer_class : Type[Any]
        The class to be used for creating layers.
    clusterable_vectors : np.ndarray
        The vectors that can be clustered.
    embedding_vectors : np.ndarray
        The embedding vectors corresponding to the clusterable vectors.
    min_clusters : int, optional
        The minimum number of clusters to form in a layer (default is 6).
    min_samples : int, optional
        The minimum number of samples for hdbscan style clustering (default is 5).
    base_min_cluster_size : int, optional
        The base minimum size of clusters for the most fine-grained cluster layer (default is 10).
    base_n_clusters : Optional[int], optional
        The base number of clusters for the most fine-grained cluster layer (default is None).
        If None then base_min_cluster_size is used; otherwise this value will override base_min_cluster_size.
    next_cluster_size_quantile : float, optional
        The quantile value to determine the size of the minimum cluster size for the next layer (default is 0.8).
    show_progress_bar : bool, optional
        Whether to show a progress bar (default is False).
    verbose : bool, optional
        Whether to show verbose output (default is False).
    **layer_kwargs : Any
        Any additional keyword arguments to be passed to the layer class.

    Returns
    -------
    Tuple[List[Any], Dict[Tuple[int, int], List[Tuple[int, int]]]]
        A tuple containing a list of created layers and a dictionary representing the cluster tree.
    """
    cluster_labels = build_raw_cluster_layers(
        clusterable_vectors,
        min_clusters=min_clusters,
        min_samples=min_samples,
        base_min_cluster_size=base_min_cluster_size,
        base_n_clusters=base_n_clusters,
        next_cluster_size_quantile=next_cluster_size_quantile,
        verbose=verbose,
    )
    cluster_tree = build_cluster_tree(cluster_labels)
    layers = [
        layer_class(
            labels,
            centroids_from_labels(labels, embedding_vectors),
            layer_id=i,
            show_progress_bar=show_progress_bar,
            **layer_kwargs,
        )
        for i, labels in enumerate(cluster_labels)
    ]
    return layers, cluster_tree


class Clusterer(ABC):

    def __init__(self):
        pass

    @abstractmethod
    def fit(
        self,
        clusterable_vectors: np.ndarray,
        embedding_vectors: np.ndarray,
        layer_class: Type[ClusterLayer],
        **layer_kwargs,
    ):
        pass

    @abstractmethod
    def fit_predict(
        self,
        clusterable_vectors: np.ndarray,
        embedding_vectors: np.ndarray,
        layer_class: Type[ClusterLayer],
        **layer_kwargs,
    ) -> Tuple[List[ClusterLayer], ClusterTree]:
        pass


class ToponymyClusterer(Clusterer):
    """
    A class for clustering data using a layered version of HDBSCAN.

    Parameters
    ----------
    min_clusters : int, optional
        The minimum number of clusters to form in a layer (default is 6).
    min_samples : int, optional
        The minimum number of samples for hdbscan style clustering (default is 5).
    base_min_cluster_size : int, optional
        The base minimum size of clusters for the most fine-grained cluster layer (default is 10).
    base_n_clusters : Optional[int], optional
        The base number of clusters for the most fine-grained cluster layer (default is None).
        If None then base_min_cluster_size is used; otherwise this value will override base_min_cluster_size.
    next_cluster_size_quantile : float, optional
        The quantile value to determine the size of the minimum cluster size for the next layer (default is 0.8).
    verbose : bool, optional
        Whether to show verbose output (default is False).

    Attributes
    ----------
    cluster_layers_ : List[ClusterLayer]
        A list of the created cluster layers.
    cluster_tree_ : Dict[Tuple[int, int], List[Tuple[int, int]]]
        A dictionary representing the cluster tree. Keys are a tuple of (layer, cluster index) and values are lists of
        tuples representing child clusters.
    """

    def __init__(
        self,
        min_clusters: int = 6,
        min_samples: int = 5,
        base_min_cluster_size: Optional[int] = 10,
        base_n_clusters: Optional[int] = None,
        next_cluster_size_quantile: float = 0.85,
        verbose=False,
    ):
        super().__init__()
        self.min_clusters = min_clusters
        self.min_samples = min_samples
        self.base_min_cluster_size = base_min_cluster_size
        self.base_n_clusters = base_n_clusters
        self.next_cluster_size_quantile = next_cluster_size_quantile
        self.verbose = verbose

        if self.base_min_cluster_size is None and self.base_n_clusters is None:
            raise ValueError(
                "Either base_min_cluster_size or base_n_clusters must be provided."
            )

    def fit(
        self,
        clusterable_vectors: np.ndarray,
        embedding_vectors: np.ndarray,
        layer_class: Type[ClusterLayer] = ClusterLayerText,
        show_progress_bar: bool = False,
        **layer_kwargs,
    ) -> Clusterer:
        self.cluster_layers_, self.cluster_tree_ = create_cluster_layers(
            layer_class,
            clusterable_vectors=clusterable_vectors,
            embedding_vectors=embedding_vectors,
            min_clusters=self.min_clusters,
            min_samples=self.min_samples,
            base_min_cluster_size=self.base_min_cluster_size,
            base_n_clusters=self.base_n_clusters,
            next_cluster_size_quantile=self.next_cluster_size_quantile,
            show_progress_bar=show_progress_bar,
            verbose=self.verbose,
            **layer_kwargs,
        )
        return self

    def fit_predict(
        self,
        clusterable_vectors: np.ndarray,
        embedding_vectors: np.ndarray,
        layer_class: Type[ClusterLayer] = ClusterLayerText,
        show_progress_bar: bool = False,
        **layer_kwargs,
    ) -> Tuple[List[ClusterLayer], ClusterTree]:
        self.fit(
            clusterable_vectors,
            embedding_vectors,
            layer_class=layer_class,
            show_progress_bar=show_progress_bar,
            **layer_kwargs,
        )
        return self.cluster_layers_, self.cluster_tree_


class KMeansClusterer(Clusterer):
    """
    A class for clustering data in layers using KMeans. This class is mostly to demonstrate how one might write
    an alternative Clusterer to the ToponymyClusterer (which uses a variation of HDBCSCAN). We recommend using the
    ToponymyClusterer in practice.

    Parameters
    ----------
    min_clusters : int, optional
        The minimum number of clusters to form in a layer (default is 6).
    base_n_clusters : int, optional
        The initial number of clusters for the most fine-grained cluster layer (default is 1024).

    Attributes
    ----------
    cluster_layers_ : List[ClusterLayer]
        A list of the created cluster layers.
    cluster_tree_ : Dict[Tuple[int, int], List[Tuple[int, int]]]
        A dictionary representing the cluster tree. Keys are a tuple of (layer, cluster index) and values are lists of
        tuples representing child clusters.
    """

    def __init__(
        self, min_clusters: int = 6, base_n_clusters: int = 1024, verbose=False
    ):
        super().__init__()
        self.min_clusters = min_clusters
        self.base_n_clusters = base_n_clusters
        self.verbose = verbose

    def fit(
        self,
        clusterable_vectors: np.ndarray,
        embedding_vectors: np.ndarray,
        layer_class: Type[ClusterLayer] = ClusterLayerText,
        show_progress_bar: bool = False,
        **layer_kwargs,
    ) -> Clusterer:
        n_clusters = self.base_n_clusters
        cluster_label_layers = []

        while n_clusters >= self.min_clusters:
            if self.verbose:
                print(f"Layer {len(cluster_label_layers)} found {n_clusters} clusters")
            kmeans = KMeans(n_clusters=n_clusters)
            cluster_labels = kmeans.fit_predict(clusterable_vectors)
            cluster_label_layers.append(cluster_labels)
            n_clusters //= 4

        self.cluster_tree_ = build_cluster_tree(cluster_label_layers)
        self.cluster_layers_ = [
            layer_class(
                labels,
                centroids_from_labels(labels, embedding_vectors),
                layer_id=i,
                show_progress_bar=show_progress_bar,
                **layer_kwargs,
            )
            for i, labels in enumerate(cluster_label_layers)
        ]
        return self

    def fit_predict(
        self,
        clusterable_vectors: np.ndarray,
        embedding_vectors: np.ndarray,
        layer_class: Type[ClusterLayer] = ClusterLayerText,
        show_progress_bar: bool = False,
        **layer_kwargs,
    ) -> Tuple[List[ClusterLayer], ClusterTree]:
        self.fit(clusterable_vectors, embedding_vectors, layer_class=layer_class, show_progress_bar=show_progress_bar, **layer_kwargs)
        return self.cluster_layers_, self.cluster_tree_


try:
    import evoc

    class EVoCClusterer(Clusterer):

        def __init__(
            self,
            min_clusters: int = 4,          
            base_min_cluster_size: Optional[int] = 10,
            base_n_clusters: Optional[int] = None,
            noise_level: float = 0.5,
            n_neighbors: int = 15,
            min_samples: int = 5,
            next_cluster_size_quantile: float = 0.85,
            n_epochs: int = 50,
            node_embedding_init: str = "label_prop",
            symmetrize_graph: bool = True,
            node_embedding_dim: Optional[int] = None,
            neighbor_scale: float = 1.0,
        ):
            super().__init__()

            self.min_clusters = min_clusters
            self.base_min_cluster_size = base_min_cluster_size
            self.base_n_clusters = base_n_clusters
            self.noise_level = noise_level
            self.n_neighbors = n_neighbors
            self.min_samples = min_samples
            self.next_cluster_size_quantile = next_cluster_size_quantile
            self.n_epochs = n_epochs
            self.node_embedding_init = node_embedding_init
            self.symmetrize_graph = symmetrize_graph
            self.node_embedding_dim = node_embedding_dim
            self.neighbor_scale = neighbor_scale

            self.evoc = evoc.EVoC(
                noise_level=noise_level,
                base_min_cluster_size=base_min_cluster_size,
                base_n_clusters=base_n_clusters,
                min_num_clusters=min_clusters,
                n_neighbors=n_neighbors,
                min_samples=min_samples,
                next_cluster_size_quantile=next_cluster_size_quantile,
                n_epochs=n_epochs,
                node_embedding_init=node_embedding_init,
                symmetrize_graph=symmetrize_graph,
                node_embedding_dim=node_embedding_dim,
                neighbor_scale=neighbor_scale,
            )

        def fit(
            self,
            clusterable_vectors: np.ndarray,
            embedding_vectors: np.ndarray,
            layer_class: Type[ClusterLayer] = ClusterLayerText,
            show_progress_bar: bool = False,
        ) -> Clusterer:
            self.evoc.fit(embedding_vectors)
            cluster_labels = self.evoc.cluster_layers_
            self.cluster_tree_ = build_cluster_tree(cluster_labels)
            self.cluster_layers_ = [
                layer_class(
                    labels,
                    centroids_from_labels(labels, embedding_vectors),
                    layer_id=i,
                    show_progress_bar=show_progress_bar,
                )
                for i, labels in enumerate(cluster_labels)
            ]
            return self

        def fit_predict(
            self,
            clusterable_vectors: np.ndarray,
            embedding_vectors: np.ndarray,
            layer_class: Type[ClusterLayer] = ClusterLayerText,
        ) -> Tuple[List[ClusterLayer], ClusterTree]:
            self.fit(clusterable_vectors, embedding_vectors, layer_class=layer_class)
            return self.cluster_layers_, self.cluster_tree_

except ImportError:
    pass



from toponymy.clustering import ToponymyClusterer, Clusterer
from toponymy.keyphrases import KeyphraseBuilder
from toponymy.cluster_layer import ClusterLayer, ClusterLayerText
from toponymy.topic_tree import TopicTree
from toponymy.llm_wrappers import LLMWrapper

from sentence_transformers import SentenceTransformer
from sklearn.utils.validation import check_is_fitted
import numpy as np

from tqdm.auto import tqdm

from typing import List, Any, Type


class Toponymy:
    """
    A class for generating topic names for vector based topic modeling.

    Parameters:
    -----------
    llm_wrapper: class
        A llm_wrapper class.  These should be objects that inherit from the LlmWrapper base classs.
    embedding_model: callable
        a function with an encode used to vectorize the objects that we are topic modeling.
    clusterer: Clusterer
        The clusterer to use for clustering the objects. This should be a clusterer that inherits from the Clusterer base class.
    layer_class: Type[Any]
        The class to used for creating layers from our objects. Default is ClusterLayerText.
    keyphrase_builder: KeyphraseBuilder
        The keyphrase builder to use for building keyphrases from the objects.
    object_description: str
        A description of the objects being topic modeled.
    corpus_description: str
        A description of the collection of objects being topic modeled.
    lowest_detail_level: float
        The lowest detail level to use for the topic names. This should be a value between 0 (finest grained detail) and 1 (very high level).
    highest_detail_level: float
        The highest detail level to use for the topic names. This should be a value between 0 (finest grained detail) and 1 (very high level).
    exemplar_delimiters: List[str]
        A list of strings that represent the delimiters for the exemplar texts. Default is ["    *\"", "\"\n"].
    show_progress_bars: bool
        Whether to show progress bars or not.

    Attributes:
    -----------
    llm_wrapper: class
        A llm_wrapper class.  These should be objects that inherit from the LlmWrapper base classs.
    embedding_model: callable
        a function with an encode used to vectorize the objects that we are topic modeling.
    clusterer: Clusterer
        The clusterer to use for clustering the objects. This should be a clusterer that inherits from the Clusterer base class.
    layer_class: Type[Any]
        The class to used for creating layers from our objects.
    keyphrase_builder: KeyphraseBuilder
        The keyphrase builder to use for building keyphrases from the objects.
    object_description: str
        A description of the objects being topic modeled.
    corpus_description: str
        A description of the collection of objects being topic modeled.
    lowest_detail_level: float
        The lowest detail level to use for the topic names. This should be a value between 0 (finest grained detail) and 1 (very high level).
    highest_detail_level: float
        The highest detail level to use for the topic names. This should be a value between 0 (finest grained detail) and 1 (very high level).
    exemplar_delimiters: List[str]
        A list of strings that represent the delimiters for the exemplar texts.
    show_progress_bars: bool
        Whether to show progress bars or not.
    clusterable_vectors_: np.array
        A numpy array of shape=(number_of_objects, clustering_dimension) used for clustering.
    embedding_vectors_: np.array
        A numpy array of shape=(number_of_objects, embedding_dimension) used for vectorizing.
    cluster_layers_: List[ClusterLayer]
        A list of ClusterLayer objects that represent the layers of the topic model.
    cluster_tree_: dict
        A dictionary that represents the tree of clusters.
    object_x_keyphrase_matrix_: np.array
        A numpy array of shape=(number_of_objects, number_of_keyphrases) that represents the objects and their keyphrases.
    keyphrase_list_: List[str]
        A list of keyphrases.
    keyphrase_vectors_: np.array
        A numpy array of shape=(number_of_keyphrases, embedding_dimension) that represents the keyphrase vectors.
    topic_names_: List[List[str]]
        A list of lists of strings that represent the topic names at each layer of the topic model.
    topic_name_vectors_: List[np.array]
        A list of numpy arrays of shape=(number_of_topics, embedding_dimension) that represent the topic names of each object
        at each layer of the topic model.

    """

    def __init__(
        self,
        llm_wrapper: LLMWrapper,
        text_embedding_model: SentenceTransformer,
        clusterer: Clusterer = ToponymyClusterer(),
        layer_class: Type[ClusterLayer] = ClusterLayerText,
        keyphrase_builder: KeyphraseBuilder = KeyphraseBuilder(),
        object_description: str = "objects",
        corpus_description: str = "collection of objects",
        lowest_detail_level: float = 0.0,
        highest_detail_level: float = 1.0,
        exemplar_delimiters: List[str] = ['    * "', '"\n'],
        show_progress_bars: bool = True,
    ):
        self.llm_wrapper = llm_wrapper
        self.embedding_model = text_embedding_model
        self.layer_class = layer_class
        self.clusterer = clusterer
        self.keyphrase_builder = keyphrase_builder
        self.object_description = object_description
        self.corpus_description = corpus_description
        self.lowest_detail_level = lowest_detail_level
        self.highest_detail_level = highest_detail_level
        self.exemplar_delimiters = exemplar_delimiters
        self.show_progress_bars = show_progress_bars

    def fit(
        self,
        objects: List[Any],
        embedding_vectors: np.array,
        clusterable_vectors: np.array,
        exemplar_method: str = "central",
        keyphrase_method: str = "information_weighted",
        subtopic_method: str = "central",
    ):
        """
        Vectorizes using the classes embedding_model and constructs a low dimension data map with UMAP if object_vectors and object_map aren't spec.

        Parameters:
        -----------
        objects: Object
            The objects over which to perform topic modeling.  These are often text documents or images.
        embedding_vectors: np.array
            An numpy array of shape=(number_of_objects, embedding_dimension) created with the same embedding_model specified in the constructor.
        clusterable_vectors: np.array
            A numpy array of shape=(number_of_objects, clustering_dimension).  It is recommended that the clustering_dimension should be low enough
            for density based clustering to be efficient (2-25).
        exemplar_method: str, Optional
            The method to use for generating exemplars. Default is "central". Other options are to be implemented.
        keyphrase_method: str, Optional
            The method to use for generating keyphrases. Default is "information_weighted". Other options are "central" and "bm25".
        subtopic_method: str, Optional
            The method to use for generating subtopics. Default is "central". Other options are "information_weighted".

        Returns:
        --------
        self: object
            Returns the instance of the class.
        """
        self.clusterable_vectors_ = clusterable_vectors
        self.embedding_vectors_ = embedding_vectors

        # Build our layers and cluster tree
        if hasattr(self.clusterer, "cluster_layers_") and hasattr(
            self.clusterer, "cluster_tree_"
        ):
            # If the clusterer has already been fit, we can skip this step
            self.cluster_layers_ = self.clusterer.cluster_layers_
            self.cluster_tree_ = self.clusterer.cluster_tree_
        else:
            self.cluster_layers_, self.cluster_tree_ = self.clusterer.fit_predict(
                clusterable_vectors,
                embedding_vectors,
                self.layer_class,
                show_progress_bar=self.show_progress_bars,
                exemplar_delimiters=self.exemplar_delimiters,
                prompt_format=(
                    "system_user"
                    if self.llm_wrapper.supports_system_prompts
                    else "combined"
                ),
            )

        # Initialize other data structures
        self.topic_names_ = [[]] * len(self.cluster_layers_)
        self.topic_name_vectors_ = [np.array([])] * len(self.cluster_layers_)
        detail_levels = np.linspace(
            self.lowest_detail_level,
            self.highest_detail_level,
            len(self.cluster_layers_),
        )

        # Get exemplars for layer 0 first and build keyphrase matrix
        if (
            hasattr(self.cluster_layers_[0], "object_to_text_function")
            and self.cluster_layers_[0].object_to_text_function is not None
        ):
            # Non-text objects: use exemplars to build keyphrase matrix
            exemplars, exemplar_indices = self.cluster_layers_[0].make_exemplar_texts(
                objects,
                embedding_vectors,
            )

            # Create aligned text list
            aligned_texts = [""] * len(objects)  # Empty strings for non-exemplars
            for cluster_idx, cluster_exemplars in enumerate(exemplars):
                for exemplar_idx, exemplar_text in zip(
                    exemplar_indices[cluster_idx], cluster_exemplars
                ):
                    aligned_texts[exemplar_idx] = exemplar_text

            # Build keyphrase matrix from aligned texts
            (
                self.object_x_keyphrase_matrix_,
                self.keyphrase_list_,
                self.keyphrase_vectors_,
            ) = self.keyphrase_builder.fit_transform(aligned_texts)
        else:
            # Text objects: build keyphrase matrix directly from objects
            (
                self.object_x_keyphrase_matrix_,
                self.keyphrase_list_,
                self.keyphrase_vectors_,
            ) = self.keyphrase_builder.fit_transform(objects)
            # Still need to generate exemplars for layer 0
            self.cluster_layers_[0].make_exemplar_texts(
                objects,
                embedding_vectors,
            )

        if self.keyphrase_vectors_ is None:
            # If the keyphrase vectors are None, we need to generate them
            self.keyphrase_vectors_ = self.embedding_model.encode(
                self.keyphrase_list_,
                show_progress_bar=self.show_progress_bars,
            )

        # Iterate through the layers and build the topic names
        for i, layer in tqdm(
            enumerate(self.cluster_layers_),
            desc=f"Building topic names by layer",
            disable=not self.show_progress_bars,
            total=len(self.cluster_layers_),
            unit="layer",
        ):
            if i > 0:  # Skip layer 0 exemplars as we already did them
                layer.make_exemplar_texts(
                    objects,
                    embedding_vectors,
                    method=exemplar_method,
                )

            layer.make_keyphrases(
                self.keyphrase_list_,
                self.object_x_keyphrase_matrix_,
                self.keyphrase_vectors_,
                self.embedding_model,
                method=keyphrase_method,
            )

            if i > 0:
                if not hasattr(self.cluster_layers_[0], "topic_name_embeddings"):
                    self.cluster_layers_[0].embed_topic_names(self.embedding_model)

                layer.make_subtopics(
                    self.topic_names_[0],
                    self.cluster_layers_[0].cluster_labels,
                    self.cluster_layers_[0].topic_name_embeddings,
                    self.embedding_model,
                    method=subtopic_method,
                )

            layer.make_prompts(
                detail_levels[i],
                self.topic_names_,
                self.object_description,
                self.corpus_description,
                self.cluster_tree_,
            )
            self.topic_names_[i] = layer.name_topics(
                self.llm_wrapper,
                detail_levels[i],
                self.topic_names_,
                self.object_description,
                self.corpus_description,
                self.cluster_tree_,
                self.embedding_model,
            )
            self.topic_name_vectors_[i] = layer.make_topic_name_vector()

        return self

    def fit_predict(
        self,
        objects: List[Any],
        object_vectors: np.array,
        clusterable_vectors: np.array,
        exemplar_method: str = "central",
        keyphrase_method: str = "information_weighted",
        subtopic_method: str = "central",
    ) -> List[np.array]:
        """
        Fit the model with objects and return the topic names.

        Parameters:
        -----------
        objects: List[Any]
            A list of objects to perform topic modeling over.
        object_vectors: np.array
            An array of shape=(number_of_objects, embedding_dimension) created with the same embedding_model specified in the constructor.
        object_map: np.array
            An array of shape=(number_of_objects, clustering_dimension).  It is recommended that the clustering_dimension should be low enough
            for density based clustering to be efficient (2-25).

        Returns:
        --------
        topic_name_vectors: List[np.array]
            A list of numpy arrays of shape=(number_of_topics, embedding_dimension) that represent the topic names of each object
            at each layer of the topic model.
        """
        self.fit(
            objects,
            object_vectors,
            clusterable_vectors,
            exemplar_method=exemplar_method,
            keyphrase_method=keyphrase_method,
            subtopic_method=subtopic_method,
        )
        return self.topic_name_vectors_

    @property
    def topic_tree_(self) -> TopicTree:
        """
        Returns the topic tree.

        Returns:
        --------
        TopicTree
            A representation of the topic tree (either html or string).
        """
        check_is_fitted(self, ["cluster_tree_", "topic_names_", "topic_name_vectors_"])

        def cluster_size(cluster_label_array):
            if cluster_label_array.min() < 0:
                return np.bincount(cluster_label_array - cluster_label_array.min())[
                    -cluster_label_array.min() :
                ].tolist()
            else:
                return np.bincount(cluster_label_array).tolist()

        topic_sizes = [
            cluster_size(layer.cluster_labels) for layer in self.cluster_layers_
        ]
        return TopicTree(
            self.cluster_tree_,
            self.topic_names_,
            topic_sizes,
            self.embedding_vectors_.shape[0],
        )



import numpy as np
from tqdm.auto import tqdm
import httpx
from tenacity import retry, stop_after_attempt, wait_exponential, wait_fixed


from typing import Optional, List

# Cohere
try:
    import cohere

    class CohereEmbedder:
        def __init__(self, api_key, model: str = "embed-multilingual-v3.0", base_url: str = None, httpx_client: Optional[httpx.Client] = None):
            self.co = cohere.ClientV2(api_key=api_key)
            self.model = model
            self.base_url = base_url
            self.httpx_client = httpx_client
            self.input_type="search_query" # We will be embedding keyphrases and subtopic names to match against documents
            self.embedding_types=['float']

        def encode(self, texts: List[str], show_progress_bar: bool = False) -> np.ndarray:
            result = []
            for i in tqdm(range(0, len(texts), 96), desc="embedding texts", disable=(not show_progress_bar)):
                response = self.co.embed(texts=texts[i:i+96], model=self.model, input_type=self.input_type, embedding_types=self.embedding_types)
                result.append(np.asarray(response.embeddings.float_))

            return np.vstack(result)
    
except ImportError:
    pass

# OpenAI
try:
    import openai

    class OpenAIEmbedder:

        def __init__(self, api_key, model: str = "text-embedding-3-small", base_url: str = None):
            self.api_key = api_key
            self.model = model
            self.base_url = base_url
            self.client = openai.OpenAI(api_key=api_key, base_url=base_url)

        def encode(self, texts: List[str], show_progress_bar: bool = False) -> np.ndarray:
            result = []
            for i in tqdm(range(0, len(texts), 96), desc="embedding texts", disable=(not show_progress_bar)):
                response = self.client.embeddings.create(texts=texts[i:i+96], model=self.model, encoding_format="float")
                result.append(np.asarray([item.embedding for item in response.data]))       

            return np.vstack(result)
        
except ImportError:
    pass

# Anthropic
try:
    import anthropic

    class AnthropicEmbedder:
        def __init__(self, api_key, model: str = "claude-3-haiku-20240307", base_url: str = None, httpx_client: Optional[httpx.Client] = None):
            self.client = anthropic.Anthropic(api_key=api_key)
            self.model = model
            self.base_url = base_url
            self.httpx_client = httpx_client

        def encode(self, texts: List[str], show_progress_bar: bool = False) -> np.ndarray:
            result = []
            for i in tqdm(range(0, len(texts), 96), desc="embedding texts", disable=(not show_progress_bar)):
                batch = texts[i:i+96]
                # Anthropic embeddings are done one at a time in the current API
                batch_embeddings = []
                for text in tqdm(batch, desc="embedding batch", disable=(not show_progress_bar), leave=False):
                    response = self.client.embeddings.create(
                        model=self.model,
                        input=text,
                    )
                    batch_embeddings.append(response.embedding)
                result.append(np.array(batch_embeddings))

            return np.vstack(result)
except ImportError:
    pass

# Microsoft Azure
try:
    import azure.ai.inference
    from azure.core.credentials import AzureKeyCredential

    class AzureAIEmbedder:
        def __init__(self, api_key: str, endpoint: str, model: str):
            self.credentials = AzureKeyCredential(api_key)
            self.client = azure.ai.inference.EmbeddingsClient(endpoint=endpoint, credential=self.credentials)
            self.model = model

        @retry(stop=stop_after_attempt(3), wait=wait_exponential(multiplier=1, min=4, max=10))
        def _encode_batch(self, texts: list) -> np.ndarray:
            # Call the Azure AI Inference API
            response = self.client.embed(
                model=self.model,
                input=[x if len(x) > 0 else "[NO_TEXT]" for x in texts]
            )
            # Extract embeddings from the response
            embeddings = [item.embedding for item in response.data]
            if len(embeddings) != len(texts):
                print(f"Warning: Expected {len(texts)} embeddings, but got {len(embeddings)}.")
                print(f"Texts: {texts}")
            assert len(embeddings) == len(texts)
            return np.array(embeddings)

        def encode(self, texts: list, show_progress_bar: bool = False) -> np.ndarray:
            result = []
            
            for i in tqdm(range(0, len(texts), 96), desc="embedding texts", disable=(not show_progress_bar)):
                embeddings = self._encode_batch(texts[i:i+96])
                result.append(embeddings)
            
            return np.vstack(result)

except ImportError as e:
    pass

# Mistral
try:
    import mistralai.client

    class MistralEmbedder:
        def __init__(self, api_key: str, model: str = "mistral-embed"):
            self.client = mistralai.client.MistralClient(api_key=api_key)
            self.model = model

        def encode(self, texts: List[str], show_progress_bar: bool = False) -> np.ndarray:
            result = []
            for i in tqdm(range(0, len(texts), 96), desc="embedding texts", disable=(not show_progress_bar)):
                response = self.client.embeddings(
                    model=self.model,
                    inputs=texts[i:i+96]
                )
                result.append(np.array([item.embedding for item in response.data]))

            return np.vstack(result)
except ImportError:
    pass

# Voyage AI
try:
    import requests

    class VoyageAIEmbedder:
        def __init__(self, api_key: str, model: str = "voyage-2"):
            self.api_key = api_key
            self.model = model
            self.base_url = "https://api.voyageai.com/v1/embeddings"
            self.headers = {
                "Authorization": f"Bearer {api_key}",
                "Content-Type": "application/json"
            }

        def encode(self, texts: List[str], show_progress_bar: bool = False) -> np.ndarray:
            result = []
            for i in tqdm(range(0, len(texts), 96), desc="embedding texts", disable=(not show_progress_bar)):
                response = requests.post(
                    self.base_url,
                    headers=self.headers,
                    json={
                        "model": self.model,
                        "input": texts[i:i+96],
                        "encoding_format": "float"
                    }
                )
                response.raise_for_status()
                data = response.json()
                result.append(np.array([item["embedding"] for item in data["data"]]))

            return np.vstack(result)
except ImportError:
    pass